#include<stdio.h>
#include<stdlib.h>
#include<time.h>

struct bst_node {
    int key;
    struct bst_node *left_child, *right_child;
} *root;

void add_node(struct bst_node **node, int number)
{
    while(*node && (*node)->key != number)
        if((*node)->key > number)
            node = &(*node)->left_child;
        else
            node = &(*node)->right_child;
    if(!*node) {
        *node = (struct bst_node *)malloc(sizeof(struct bst_node));
        if(*node) {
            (*node)->key = number;
            (*node)->left_child = (*node)->right_child = NULL;
        }
    }
}

void print_bst_inorder(struct bst_node *node)
{
    if(node) {
        print_bst_inorder(node->left_child);
        printf("%4d ",node->key);
        print_bst_inorder(node->right_child);
    }
}

void print_bst_preorder(struct bst_node *node)
{
    if(node) {
        printf("%4d ",node->key);
        print_bst_preorder(node->left_child);
        print_bst_preorder(node->right_child);
    }
}

void print_bst_postorder(struct bst_node *node)
{
    if(node) {
        print_bst_postorder(node->left_child);
        print_bst_postorder(node->right_child);
        printf("%4d ",node->key);
    }
}

struct bst_node *find_minimum(struct bst_node *node)
{
    while(node && node->left_child)
        node = node->left_child;
    return node;
}

struct bst_node *find_maximum(struct bst_node *node)
{
    while(node && node->right_child)
        node = node->right_child;
    return node;
}

struct bst_node *locate(struct bst_node *node, int number)
{
    while(node && node->key != number)
        if(node->key > number)
            node = node->left_child;
        else
            node = node->right_child;
    return node;
}

void remove_bst_nodes(struct bst_node **node)
{
    if(*node) {
        remove_bst_nodes(&(*node)->left_child);
        remove_bst_nodes(&(*node)->right_child);
        free(*node);
        *node=NULL;
    }
}

unsigned int count_nodes(struct bst_node *node)
{
    if(node)
        return count_nodes(node->left_child) + 1 + count_nodes(node->right_child);
    else
        return 0;
}

struct bst_node *isolate_predecessor(struct bst_node **node)
{
    while((*node)->right_child)
        node = &(*node)->right_child;
    struct bst_node *predecessor = *node;
    *node = (*node)->left_child;
    return predecessor;
}

void delete_node(struct bst_node **node, int number)
{
    while(*node && (*node)->key != number)
        if ((*node)->key > number)
            node = &(*node)->left_child;
        else
            node = &(*node)->right_child;

    if (*node) {
        struct bst_node *node_to_delete = *node;
        if(!node_to_delete->left_child)
            *node = (*node)->right_child;
        else if(!node_to_delete->right_child)
            *node = (*node)->left_child;
        else {
            node_to_delete = isolate_predecessor(&(*node)->left_child);
            (*node)->key = node_to_delete->key;
        }
        free(node_to_delete);
    }
}

int main(void)
{
    srand(time(0));
    for(int i=0; i<10; i++)
        add_node(&root, -10 + rand() %21);
    printf("Number of nodes in the binary search tree: %u\n", count_nodes(root));
    print_bst_inorder(root);
    puts("");
    print_bst_preorder(root);
    puts("");
    print_bst_postorder(root);
    puts("");
    if(root) {
        printf("Minimum key: %d\n", find_minimum(root)->key);
        printf("Maximum key: %d\n", find_maximum(root)->key);
    }
    int number_to_delete = -10+rand()%21;
    printf("Key to be deleted: %d\n", number_to_delete);
    struct bst_node *result = locate(root, number_to_delete);
    if(result)
        printf("The key is in the binary search tree: %d\n", result->key);
    else
        puts("The key is not in the binary search tree.");
    delete_node(&root, number_to_delete);
    printf("The BST after trying to delete the key %d.\n",number_to_delete);
    printf("Number of nodes in the binary search tree: %u\n", count_nodes(root));
    puts("Keys in the BST.");
    print_bst_inorder(root);
    puts("");
    remove_bst_nodes(&root);
    return 0;
}
